require 'rubygems'

require 'net/http'
require 'net/https'
require 'uri'
require 'json'


module Backrack; end

class RPCError < RuntimeError
  attr_accessor :name 
  attr_accessor :message
  attr_accessor :code
  def to_str
    "#{self.name}: #{self.message}"
  end
end


# Backrack client
#
# Example usage
# <pre>
#   require 'backrack/client'
#   client = Backrack::Client.new('https://example.com','john','pass')
#   Foo = client.get_class 'Foo'
# </pre>
#
# Or:
# <pre>
#  require 'backrack/client'
#  class Foo < Backrack::RPCProxy
#    connect_to "http://example.com:5000"
#  end 
# </pre>
#
# With the second method if you want to define method stubs inside the class,
# remember to put host method at the end. Otherwise stubs will overwrite real methods


class Backrack::Client
  def initialize(url, http_login=nil, http_password=nil)
    @remote_url = url
    @remote_url += '/' unless @remote_url[-1].chr == '/'
    @remote_url = URI.parse @remote_url
    
    if http_login and http_password  
      @remote_url.user = http_login
      @remote_url.password = http_password
    end
  end
  
  def connect()
    http = Net::HTTP.new(@remote_url.host, @remote_url.port)
    http.use_ssl=true if @remote_url.port == 443    
    http.read_timeout= 300
    http.start
  end
  
  def backrack_call(klass, object_id, method, params=[])
    method_url = object_id ? "/#{klass}/#{object_id}" : "/#{klass}"
    
    remote = connect
    
    headers = {
      "accept" => "*/*, application/json", 
      "content-type"=> "application/json",
      "Content-Type" => 'application/json',
      'HTTP_ACCEPT' => 'application/json',
    } 
    
    request = {
      'method' => method, 
      'version' => '1.0', 
      'params' => params  
    }
    
    res, body = remote.post(method_url, request.to_json, headers)
    
    case res
    when Net::HTTPSuccess, 200  
      begin
        res = JSON.load(body)
      rescue
        raise  RPCError, "Incorrect JSON! Got: #{res.body}", caller
      end
      if res['error'] 
        rpc_error = RPCError.new
        rpc_error.message = res['error']['message']
        rpc_error.code = res['error']['code']
        rpc_error.name = res['error']['name']
        raise rpc_error, res['error'].to_s, caller
      end    
      res = method.match(/^system\./) ? res : res['result']    
    else
      raise RPCError, "request was not successfull status: #{res}", caller
    end
    remote.finish
    
    return(res)
  end
  
  def get_class(name, klass=nil)
    klass ||= Class.new(Backrack::RPCProxy)
    klass.backrack_client = self
    klass.name = name
    
    k_desc = self.backrack_call(name, nil, "system.describe", [])
    singleton = k_desc["procs"].any?{|method| method["name"] == "class.instance"}
    
    text= k_desc["procs"].map do |method|
      params = method["params"].map{|i| i["name"]}.join(",")
      method_name = method["name"]
      
      next if ["class.new", "class.instance"].include? method_name
      
      if method_name["class."]
        fun = "  def #{method_name.sub(/class\./, "self.")}(#{params})\n"
        fun+= "    @@backrack_client.backrack_call(self.name, nil, \"#{method_name}\", [#{params}])\n"
      elsif singleton
        fun = "  def #{method_name}(#{params})\n"
        fun+= "    @@backrack_client.backrack_call(self.class.name, nil, \"#{method_name}\", [#{params}])\n"
      else
        fun = "  def #{method_name}(#{params})\n"
        fun+= "    @@backrack_client.backrack_call(self.class.name, @_rpc_object_id, \"#{method_name}\", [#{params}])\n"
      end
      fun+= "  end\n"
      fun
    end.join("\n")
    
    klass.class_eval(text)
    klass
  end
  
  def print_class(name)
    k_desc = self.backrack_call(name, nil, "system.describe", [])
    singleton = k_desc["procs"].any?{|method| method["name"] == "class.instance"}
    
    text = "class #{name} < Backrack::RPCProxy\n\n"
    text+= "  include Singleton\n\n" if singleton
    text+= k_desc["procs"].map do |method|
      params = method["params"].map{|i| i["name"]}.join(",")
      method_name = method["name"]
      
      next if ["class.new", "class.instance"].include?(method_name)
      
      if method_name["class."]
        fun = "  def #{method_name.sub(/class\./, "self.")}(#{params})\n"
        fun+= "    @@backrack_client.backrack_call(self.name, nil, \"#{method_name}\", [#{params}])\n"
      elsif singleton
        fun = "  def #{method_name}(#{params})\n"
        fun+= "    @@backrack_client.backrack_call(self.class.name, nil, \"#{method_name}\", [#{params}])\n"
      else
        fun = "  def #{method_name}(#{params})\n"
        fun+= "    @@backrack_client.backrack_call(self.class.name, @_rpc_object_id, \"#{method_name}\", [#{params}])\n"
      end
      fun+= "  end\n"
      fun
    end.join("\n")
    text+= "end\n"
    text+= "\n #{name}.connect_to \"put.your.host.here.com\"\n"
    
    text
  end
end

class Backrack::RPCProxy
  attr_accessor :_rpc_object_id
  
  class << self
    alias local_new new
    protected :local_new
    attr_accessor :name
    
    def backrack_client=(c)
      @@backrack_client = c
    end
    
    def new(*params)
      o = @@backrack_client.backrack_call(self.name, nil, "new", params)
      
      object = local_new()
      object._rpc_object_id = o["object_id"]
      object
    end
    
    def connect_to(url, login=nil, password=nil)
      @@backrack_client = Backrack::Client.new(url, login, password)
      @@backrack_client.get_class(self.name, self)
    end
  end
end
  
  
