require 'digest/md5'
require 'date'
require 'thread'

autoload :Session, "backrack/session"

class WrongRequest < RuntimeError; end

class Backrack::Dispatcher
  
  SPECIAL_PARAMS = ["session", "params", "response", "request"]
  
  FORBIDDEN_METHODS = SPECIAL_PARAMS + ["superclass", "allocate", "initialize",
    "instance_methods_ex", "modules", "to_yaml", "yaml_tag_subclasses?", 
    "json_creatable?", "_dump"
  ].uniq.freeze
  
  attr_reader :class_name
  attr_reader :method_name
  
  class << self
    def init!
      @@mime_types||={}
    end
    
    def mime_types
      @@mime_types
    end
  
    def register(type, o)
      case type
      when Array
        type.each do |t|
          t = ::MIMEParse.parse_mime_type(t.to_s)
          Backrack.logger.debug "registering serializer: #{t.inspect}"
          @@mime_types[t]=o
        end
      when String
        type = ::MIMEParse.parse_mime_type(type.to_s)
        Backrack.logger.debug "registering serializer: #{type.inspect}"
        @@mime_types[type]=o
      end
    end
      
    def call(env)
      if Backrack.config[:env]=="development"
        Backrack.discover(true)
        reloader = File.join(Backrack.config[:app_root], "config", "reload.rb")
        load(reloader) if File.exists?(reloader)
      end      
      
      # ignore favicons :)
      return([404, {}, ""]) if env["path_info"] =~ /favicon/i
      
      #spawn new instance, and process the call :)
      Thread.new do
        dispatcher = self.new
        Thread.current[:dispatcher] = dispatcher
        dispatcher.call(env)
      end.value
    end
  end

  def set_thread_local_variables
    [:env, :request, :response, :session].each do |var|
      Thread.current[var] = instance_variable_get("@#{var}")
    end
  end
  
  # process the call
  def call(env)
    @env = env
    @request = Rack::Request.new(env)
    @response = Rack::Response.new
    @session = nil

    Backrack.logger.info "\033[32;1m Started processing #{@request.path_info} \033[0m"
    # we check if we support this class before we do anything else
    @class_name = @request.path_info.split("/")[1]
    @class_name &&= @class_name.camelize
    if @class_name && !Backrack::ObjectManager.supported?(@class_name)
      Backrack.logger.debug "'#{@class_name}' is not supported by this server"
      @response.status = 404
      return @response.finish
    end
    
    begin 
      get_session if Backrack.config[:use_session]
      set_thread_local_variables
      choose_serializer
      prepare_request
      get_object
      params = get_params
      result = call_method(params)
      Backrack.logger.info "\033[32;1m Finished processing #{@request.path_info} \033[0m"     
      if result.kind_of?(Rack::Response)
        return(result.finish)
      else
        @response.write(result.to_s) 
      end
    rescue WrongRequest => e
      @response.status = 404
    end
    @response.finish
  end
  
  # Get session information from cookie
  def get_session
    Backrack.logger.debug "geting session from cookie"
    unless cookie_hash = @request.cookies["backrack_user_hash"]
      cookie_hash = Digest::MD5.hexdigest("backrack_"+Time.now.to_s)
      @response.set_cookie("backrack_user_hash", {
          :path => "/", 
          :value => cookie_hash, 
          :expires => (Time.now+Backrack.config[:session_expiration_time].to_i),
        })
    end
    @session_key = cookie_hash
    @session = Session.instance(@session_key)
  end
  
  # choose serializer to use based on the mime-type of http request
  def choose_serializer
      
    @request.path_info =~ /\.(\w{1,4})\//
    file_ext = $1
    
    # File extensions override Accept:-headers.
    if exttype = Rack::Mime::MIME_TYPES[file_ext]
      @env["HTTP_ACCEPT"] = exttype
    end

    @env["HTTP_ACCEPT"] ||= "text/plain"
    
    raise if @@mime_types.keys.empty?

    Backrack.logger.debug(@env["HTTP_ACCEPT"].inspect)

    match = ::MIMEParse.best_match(@@mime_types.keys, @env["HTTP_ACCEPT"])

    if match
      @serializer = @@mime_types[match].new
    else
      @response.status = 406          # Not Acceptable
      raise RuntimeError, "could not find acceptable serializer, this shouldn't happen :]", caller
    end
    
    Backrack.logger.info "using serializer #{@serializer.class.name} for #{@env["HTTP_ACCEPT"]}"
  end
  
  # prepares information needed to process rpc request.
  # if serializer responds to :process_rpc_request it's called, if not generic
  # method is used
  def prepare_request
    Backrack.logger.debug "checking and prepareing request"
    rpc_request = if @serializer.respond_to?(:process_rpc_request)
      payload = @request.post? ? @request.env["rack.input"].read : @request.params
      @serializer.process_rpc_request(@request.path_info, payload, @request.post?)
    else
      self.process_rpc_request(@request)
    end
      
    @class_name, @instance_id, @method_name, @params = *rpc_request
  end
  
  # standard request processing method, chooses class name, object id,
  # method name to call based on url as well as params to pass to that method
  #
  # * / => DefaultClass.get (if DefaultClass is given and request method is get)
  # * /Klass => Klass.post (if request_method is post)
  # * /Klass/foo?a=b => Klass.foo("b") 
  # * /Klass/1 => Klass.send(:'1') 
  # * /Klass/1/ => object_with_class_Klass_and_id_1.get (if request method is get)
  # * /Klass/1/foo => object_with_class_Klass_ans_id_1.foo
  # * /Klass/1/foo/bar => object_with_class_Klass_ans_id_1.foo("bar")
  def process_rpc_request(request)
    path = request.path_info.split("/", -1)[1..-1] # we get path without first empty element
    
    params = request.params # we get params
    
    class_name = path && path.shift # we get class
    raise(WrongRequest) unless class_name || Backrack.default_class 
    
    if path[1]
      instance_id, method_name = path.shift, path.shift
    else
      method_name = path.shift
      method_name = nil if method_name && method_name.empty?
    end

    params ||= {}
    result = if instance_id && method_name 
      #/Klass/-141124/method or /Klass/-1234/
      method_name = request.request_method.downcase if method_name.empty?
      # /Klass/-1423/method/param1/param2    
      params = path if params.empty? && !path.empty?
      instance_id = instance_id.to_i if /^-?\d+$/.match(instance_id)
      [class_name, instance_id, method_name, params ]
    elsif method_name && (method_name =~ /^-?\d+$/)
      # /Klass/-666
      [class_name, method_name.to_i, request.request_method.downcase, params ]
    elsif method_name
      # /Klass/method
      [class_name, nil, method_name, params ]
    elsif class_name && !class_name.empty?
      # /Klass/get  (or post, or delete etc.)
      [class_name, nil, request.request_method.downcase, params ]        
    else
      # /DefaultClass/get  (or post, or delete etc.)
      [Backrack.default_class, nil, request.request_method.downcase, params]
    end
    
    Backrack.logger.debug result.map(&:inspect).join("/") 
    
    return(result)
  end
  
  # finds the object matching criteria that were selected during process_rpc_request call
  def get_object
    Backrack.logger.debug "preparing the object"
    
    @class = Backrack::ObjectManager.get_class(@class_name)

    # we check if serializer have any special methods on it's own
    @serializer_method = 
      @serializer.respond_to?(:special_method?) && @serializer.special_method?(@method_name)
    
    # we check in which mode we operate, class mode or normal mode
    if @class.respond_to? :instance
      @class_mode = !!(@method_name =~ /^class\./)
    else
      @class_mode = @instance_id.nil?
    end
    @method_name = @method_name.split(".")[-1]

    Backrack.logger.debug "in #{@class_mode ? "class" : "instance"} mode"
    
    # we get @class/@object pair
    if @class_mode
      @object = @class
      @class = @class.metaclass 
      allowed = @class.instance_methods - Class.instance_methods - FORBIDDEN_METHODS + ["new"]
      check_if_allowed(allowed)
    else
      allowed = @class.instance_methods - Object.instance_methods + 
        @class.instance_methods(false) - FORBIDDEN_METHODS
      check_if_allowed(allowed)
      @object = Backrack::ObjectManager.instance(@session_key, @class, @instance_id)
    end
    [@class, @object, @method_name]
  end
  
  # check if particular method is allowed on object
  def check_if_allowed(allowed)
    unless allowed.include?(@method_name) || @serializer_method
      Backrack.logger.warn("calling #{@method_name}, on #{@object.to_s} is not allowed\n"+
          "allowed methods: #{allowed.inspect}")
      raise(WrongRequest, "calling #{@method_name}, on #{@object.to_s} is not allowed") 
    end
  end
  
  # parameters normalization stolen from merb (i think)
  # makes params look like this: {:raz => {:dwa => 1, :trzy => 2}} 
  def normalize_params(parms, name, val=nil)
    name =~ %r([\[\]]*([^\[\]]+)\]*)
    key = $1 || ''
    after = $' || ''

    if after == ""
      parms[key] = val
    elsif after == "[]"
      (parms[key] ||= []) << val
    elsif after =~ %r(^\[\])
      parms[key] ||= []
      parms[key] << normalize_params({}, after, val)
    else
      parms[key] ||= {}
      parms[key] = normalize_params(parms[key], after, val)
    end
    parms
  end

  # normalizes params and prepares named params to be passed to method call.
  # alos it adds 4 special params :) and sets them in serializer instance
  def get_params
    Backrack.logger.debug "geting params from request"
    special_params = {}
    SPECIAL_PARAMS.each{|k| 
      value = self.instance_variable_get("@#{k}")
      special_params[k.to_sym] = value
      @serializer.send("#{k}=", value)
    }
    
    method = if @serializer_method
      @serializer.class.instance_method(@method_name)
    elsif @class_mode && @method_name == "new"
      @object.instance_method("initialize")
    else
      @object.method(@method_name)      
    end
    
    if @params.kind_of?(Hash)
      normalized_params = {}
      @params.each_pair do |k,v|
        normalize_params(normalized_params, k, v)
      end
      @params.replace(normalized_params)
    end
      
    final_params = case @params  
    when Hash
      params = @params.merge(special_params)
      method.params.map{|el| params[el] || params[el.to_s]} 
    when Array
      params_list = @params
      method.params.each_with_index do |name, i|
        params_list.insert(i, special_params[name]) if special_params.keys.include?(name)
      end
      a = method.arity # lets check arity
      # we make sure that method get all parameters she can handle
      a < 0 ? params_list : a == 0 ? [] : params_list[0..a-1] # ugly but hey it;s 1:30AM!! 
    else
      raise(RuntimeError, "wrong type of params #{@params.class} instead of Array or Hash")
    end         
    
    return(final_params)
  end
  
  # finally lets call that method, if serializer have speciall method :run
  # let's call it, if not we use normal send
  def call_method(request_params)
    if @serializer_method
      Backrack.logger.debug "calling special method: #{@method_name} on #{@serializer.class}"+
        " with params #{request_params.inspect}"
      request_params.unshift(@object)
      
      result = @serializer.send(@method_name, *request_params)
    else
      msg = @object.class == Class ? "class #{@object.name}" : "object of class #{@object.class.to_s}"
      Backrack.logger.debug "calling method: #{@method_name} on #{msg}"+
        " with params #{request_params.inspect}"
    
      if @serializer.respond_to?(:run)
        result = @serializer.run(@object, @method_name, request_params)
      else
        result = @object.send(@method_name, *request_params)
      end
    end
    return(result)
  end
  
  init!
end
