# Severly modified version, that handles stuff good :)

module MIMEParse
  module_function

  # Carves up a mime_type and returns an Array of the
  #  [type, subtype, params] where "params" is a Hash of all
  #  the parameters for the media range.
  #
  # For example, the media range "application/xhtml;q=0.5" would
  #  get parsed into:
  #
  # ["application", "xhtml", { "q" => "0.5" }]
  def parse_mime_type(mime_type)
    parts = mime_type.split(";")
    params = {}

    parts[1..-1].map do |param|
      k,v = param.split("=").map { |s| s.strip }
      params[k] = v
    end

    type, subtype = parts[0].split("/")
    raise "malformed mime type" unless subtype

    unless params.has_key?("q") and params["q"] and params["q"].to_f and params["q"].to_f <= 1 and params["q"].to_f >= 0
      params["q"] = "1"
    end
    [type.strip, subtype.strip, params]
  end
  
  # Returns the quality "q" of a mime_type when compared against
  #  the media-ranges in ranges. For example:
  #
  #     irb> quality("text/html", "text/*;q=0.3, text/html;q=0.7, text/html;level=1, text/html;level=2;q=0.4, */*;q=0.5")
  #     => 0.7
  
  # Find the best match for a givven mime_type against a list of
  #  media_ranges that have already been parsed by #parse_media_range
  #
  # Returns the "q" quality parameter of the best match, 0 if no match
  #  was found. This function behaves the same as #quality except that
  #  "parsed_ranges" must be an Enumerable of parsed media ranges.
  def match_quality(mime_type, acceptable)
    best_fitness = -1
    target_type, target_subtype, target_params = mime_type
    
    acceptable.each do |type, subtype, params|
      param_matches = target_params.find_all{ |k,v| k != "q" and params.has_key?(k) and v == params[k] }.length

      if (type == target_type or type == "*" or target_type == "*") and
          (subtype == target_subtype or subtype == "*" or target_subtype == "*")
        fitness = ((type == target_type) && (subtype == target_subtype)) ? 100 : 0 # direct hit on type and subtype
        fitness += ((type == target_type) && (type != "*")) ? 100 : 0 # direct gets 100 points
        fitness += ((subtype == target_subtype) && (subtype != "*")) ? 25 : 0 # secondary kit gets 25 points
        fitness += (params["q"].to_f * 10).to_i # what would client want more?
        fitness += (target_params["q"].to_f * 5).to_i # what we would like to serv more?
        fitness += param_matches # additional stuff
        
        if fitness > best_fitness
          best_fitness = fitness
        end
      end
    end

    best_fitness.to_f
  end

  # Takes a list of supported mime-types and finds the best match
  #  for all the media-ranges listed in header. The value of header
  #  must be a string that conforms to the format of the HTTP Accept:
  #  header. The value of supported is an Enumerable of mime-types
  #
  #     irb> best_match(["application/xbel+xml", "text/xml"], "text/*;q=0.5,*/*; q=0.1")
  #     => "text/xml"
  def best_match(supported, header)
    # we get an array of acceptable types here
    acceptable = header.split(",").map{|r| parse_mime_type(r) }
    if supported.kind_of?(String) 
      supported = supported.dup.split(",")
    end
    
    # we'll calculate quality of match for each supported
    supported.map! do |mime_type|
      parsed_type = mime_type.kind_of?(String) ? parse_mime_type(mime_type) : mime_type
      [match_quality(parsed_type, acceptable), mime_type]
    end
    
    # we sort by quality
    result = supported.sort_by{|s| s[0]}.last

    result[0] ? result[1] : nil
  end
end

