require 'json'

unless defined?(Backrack::JsonRPC)
  class Backrack::JsonRPC
    Backrack::Dispatcher.register(["application/json-rpc;q=1","application/json;q=0.9", "text/javascript;q=0.7"], self)
    ALLOWED_RESULTS = [Array, Hash, String, Fixnum, Symbol, TrueClass, 
      FalseClass, NilClass, Float, Date, Time, DateTime]
  
    attr_accessor :request, :response, :params, :session
  
    def special_method?(name)
      !!name["system."]
    end
  
    def wrap_result(res)
      {
        "version" => "1.0",
        "result" => res,
      }.to_json
    end
  
    def wrap_error(e)
      {
        "version" => "1.0",
        "error" => {
          "name" => "RPCError",
          "message" => [e.to_s, (e.backtrace.join("\n") rescue e.backtrace)],
          "code" => 666,
        }
      }.to_json
    end
  
    def wrap_object(o)
      Backrack::ObjectManager.register(o, @session.id)
      {
        "object_id" => o.respond_to?(:id) ? o.id : o.object_id,
        "class_name" => o.class.to_s,
        "methods" => o.methods_ex
      }
    end
  
    #  def unwrap_objects(o)
    #    case o
    #    when Array
    #      o.map{|a| unwrap_objects(a)}
    #    when Hash
    #      o = unwrap_object(o)
    #      o.kind_of?(Hash) ? o.each_pair{|k,v| o[k] = unwrap_objects(v)} : o
    #    else 
    #      o
    #    end
    #  end
    #  
    #  def unwrap_object(h)
    #    if oid = h["object_id"] && klass = h["class_name"]
    #      o = ObjectSpace._id2ref(oid.to_i) rescue nil
    #      o.class.name == klass ? o : h 
    #    else
    #      h
    #    end
    #  end
  
    def describe(object)
      host = Backrack.config[:host]
      port = Backrack.config[:port]
      klass = object.class == Class ? object : object.class  
      imethods = klass.instance_methods_ex.enum_for(:each_pair).collect do |method, parameters| 
        unless Backrack::Dispatcher::FORBIDDEN_METHODS.include?(method.to_s)
          {
            "name"    => method.to_s,
            "summary" => (klass.respond_to?(:doc) ? klass.doc(method) : "none" ),
            "help"    => "http://#{host}:#{port}/#{klass.to_s}/#{method}.html",
            "params"  => parameters.collect do |param| 
              { "name" => param.to_s, "type" => "str" } 
            end,
            "return" => {
              "type" => "str" 
            },
          }
        end
      end.compact
    
      cmethods = klass.methods_ex.enum_for(:each_pair).collect do |method, parameters| 
        # we get parameters for .new method from initialize
        parameters = klass.instance_method("initialize").params if method == 'new'         
        unless Backrack::Dispatcher::FORBIDDEN_METHODS.include?(method.to_s)
          {
            "name"    => "class.#{method}",
            "summary" => (klass.respond_to?(:doc) ? klass.doc(method) : "none" ),
            "help"    => "http://#{host}:#{port}/#{klass.to_s}/#{method}.html",
            "params"  => parameters.collect do |param| 
              { "name" => param.to_s, "type" => "str" } 
            end,
            "return" => {
              "type" => "str" 
            },
          }
        end
      end.compact  
    
      {
        "name"    => klass.to_s,
        "id"      => "urn:uuid:"+ UUID.create_sha1("http://#{host}:#{port}/#{klass.to_s}", UUID::NameSpace_URL).to_s,
        "summary" => (klass.respond_to?(:doc) ? klass.doc : klass.to_s),
        "help"    => "http://#{host}:#{port}/#{klass.to_s}/index.html",
        "address" => "http://#{host}:#{port}/#{klass.to_s}",
        "procs" => cmethods + imethods 
      }.to_json
    end
  
    def listMethods(object)
      klass = object.class == Class ? object : object.class
      imethods = klass.instance_methods.reject{|m|Backrack::Dispatcher::FORBIDDEN_METHODS.include?(m)}
      cmethods = klass.methods.reject{|m|Backrack::Dispatcher::FORBIDDEN_METHODS.include?(m)}.map{|m| "class.#{m}"}
      (cmethods + imethods).to_json
    end
  
    def echo(object, arg)
      arg.to_json
    end
  
    def process_rpc_request(path, payload, post = true)
      path = path.split("/")[1..-1]
    
      class_name = path.shift.camelize || raise(NoMethodError)
      instance_id = path.shift
    
      request_params, method_name = [], nil # scope
      if post
        rpc_request = JSON.load(payload)
        #      params = unwrap_objects(rpc_request["params"])
        request_params = rpc_request["params"]
        method_name = rpc_request["method"]
      else
        method_name = path.shift 
        unless method_name
          method_name = instance_id || "system.describe"
          instance_id = nil
        end
        request_params = payload
      end
    
      Backrack.logger.debug [class_name, instance_id, method_name, request_params ].map(&:inspect).join("/") 
      [class_name, instance_id, method_name, request_params]
    end
    
    def run(object, method_name, args)
      begin
        result = object.send(method_name, *args)
        result = ALLOWED_RESULTS.any?{|c| result.kind_of?(c)} ? result : wrap_object(result)
        wrap_result(result)
      rescue Exception => e
        wrap_error(e)
      end
    end
  end
end
